/* This is free and unencumbered software released into the public domain.
 * 
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * See UNLICENSE for complete information.
 *
 * This code implements our modified Gauss-sieve algorithm.
 * This file contains shared routines between the different implementations.
 *
 * See for a detailed description of the methods used:
 * Joppe W. Bos, Michael Naehrig, and Joop van de Pol:
 * Sieving for Shortest Vectors in Ideal Lattices:
 * a Practical Perspective. Cryptology ePrint Archive, 
 * Report 2014/880, IACR, 2014.
 * http://eprint.iacr.org/2014/880.pdf
 */

#ifndef __SHARED
#define __SHARED

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <emmintrin.h>
#include <malloc.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

int32_t N;
int32_t n;
char *fileName; //Basis fileName

#define INT int16_t
#define IP int64_t

#define setbit(a,x)   ((a)[(x)/64] |= (((uint64_t) 1) << (uint64_t) ((x)%64)))
#define getbit(a,x)   (((a)[(x)/64] >> (uint64_t) ((x)%64)) & 1)
#define clearbit(a,x) ((a)[(x)/64] &= ((~((uint64_t) 0)) - (((uint64_t) 1) << (uint64_t) ((x)%64))))
#define round(x)      ((x)>=0?(INT)((x)+0.5):(INT)((x)-0.5))

#define UPDATE_MU(new,cur) \
  if (mu == 1)       new.inprod = new.inprod - 2 * inprodnew + cur.inprod;   \
  else if (mu == -1) new.inprod = new.inprod + 2 * inprodnew + cur.inprod;   \
    
#define UPDATE_MU_STAR(new,cur) \
  if (mu == 1)       new->inprod = new->inprod - 2 * inprodnew + cur->inprod;   \
  else if (mu == -1) new->inprod = new->inprod + 2 * inprodnew + cur->inprod;   \

/* Reads the basis */
uint64_t readbasis (char *str, INT **b);

/* Converts numbers to and from SSE2 registers*/
void fromSSE2 (INT *c, __m128i *a);
void toSSE2 (__m128i *c, INT *a);

/* Routines required for sampling */
void gram_schmidt (double **gs, double **mu, double *lbj, INT **b, int64_t *goal);
void sample_init (double* lbj, double* vars);
int32_t sampleZ(double c, double var, double t);
void sample (INT *v, INT **b, double **mu, double *vars, double t);

#endif /* __SHARED */
